;; Sets up a complete central nervous system, and connects it to a breve object

(in-package :carausius-morosus)

(defclass central-nervous-system (neural-network)
  ((breve-object
    :accessor central-nervous-system-breve-object
    :documentation "A pointer to the breve wrapper."
    :initarg :breve-object
    :initform (make-instance 'carausius-morosus))
   (lobe
    :accessor central-nervous-system-lobe
    :documentation "The lobe."
    :initarg :lobe
    :initform (make-instance 'neural-network))
   (locomotion-controller
    :accessor central-nervous-system-locomotion-controller
    :documentation "The controller for coordinating/moving the legs."
    :initarg :locomotion-controller
    :initform (make-instance 'locomotion-controller))))

(defmethod initialize-instance :after ((cns central-nervous-system) &rest args)
  (declare (ignore args))
  (let ((bo (central-nervous-system-breve-object cns))
        (l1 (locomotion-controller-l1
             (central-nervous-system-locomotion-controller cns)))
        (l2 (locomotion-controller-l2
             (central-nervous-system-locomotion-controller cns)))
        (l3 (locomotion-controller-l3
             (central-nervous-system-locomotion-controller cns)))
        (r1 (locomotion-controller-r1
             (central-nervous-system-locomotion-controller cns)))
        (r2 (locomotion-controller-r2
             (central-nervous-system-locomotion-controller cns)))
        (r3 (locomotion-controller-r3
             (central-nervous-system-locomotion-controller cns))))
    ;; Connect coxa-femur angle sensors to breve object (continous firing sensors, not just on/off)
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor l1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor l1))
                    (- (* (/ (- (carausius-morosus-tcl0-angle bo) carausius-morosus::*AEP-L1*)
                             (- carausius-morosus::*PEP-L1* carausius-morosus::*AEP-L1*))
                          40)
                       20))))
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor l2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor l2))
                    (- (* (/ (- (carausius-morosus-tcl1-angle bo) carausius-morosus::*AEP-L2*)
                             (- carausius-morosus::*PEP-L2* carausius-morosus::*AEP-L2*))
                          40)
                       20))))
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor l3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor l3))
                    (- (* (/ (- (carausius-morosus-tcl2-angle bo) carausius-morosus::*AEP-L3*)
                             (- carausius-morosus::*PEP-L3* carausius-morosus::*AEP-L3*))
                          40)
                       20))))
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor r1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor r1))
                    (- (* (/ (- (carausius-morosus-tcr0-angle bo) carausius-morosus::*AEP-R1*)
                             (- carausius-morosus::*PEP-R1* carausius-morosus::*AEP-R1*))
                          40)
                       20))))
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor r2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor r2))
                    (- (* (/ (- (carausius-morosus-tcr1-angle bo) carausius-morosus::*AEP-R2*)
                             (- carausius-morosus::*PEP-R2* carausius-morosus::*AEP-R2*))
                          40)
                       20))))
    (setf (sensor-neuron-sensor-function (leg-controller-coxa-angle-sensor r3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-coxa-angle-sensor r3))
                    (- (* (/ (- (carausius-morosus-tcr2-angle bo) carausius-morosus::*AEP-R3*)
                             (- carausius-morosus::*PEP-R3* carausius-morosus::*AEP-R3*))
                          40)
                       20))))
     ;; Connect femur-tibia neurons to breve object
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron l1))
          #'(lambda (firing-frequency) 
(abs (setf (carausius-morosus-ftl0-down bo) ; NOTE: Must use abs, to measure force used!
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron l2))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-ftl1-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron l3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-ftl2-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron r1))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-ftr0-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron r2))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-ftr1-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-femur-tibia-neuron r3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-ftr2-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-femur-tibia-force* 2)))))
    ;; Connect foot neurons to breve object
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron l1))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfl0-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron l2))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfl1-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron l3))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfl2-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron r1))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfr0-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron r2))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfr1-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    (setf (motor-neuron-motor-function (leg-controller-foot-neuron r3))
          #'(lambda (firing-frequency) (abs (setf (carausius-morosus-cfr2-down bo)
                                                  (* (- firing-frequency 0.5) *maximum-leg-down-force* 2)))))
    ;; Connect forward angle sensors (using a func that turns on or off
    ;; external current depending on sensor)
    (setf (sensor-neuron-sensor-function
           (leg-controller-forward-angle-sensor l1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor l1))
                    (funcall
                     (if (<= (carausius-morosus-tcl0-angle bo) *AEP-L1*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor l1))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-forward-angle-sensor l2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor l2))
                    (funcall
                     (if (<= (carausius-morosus-tcl1-angle bo) *AEP-L2*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor l2))))))
    (setf (sensor-neuron-sensor-function 
           (leg-controller-forward-angle-sensor l3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor l3))
                    (funcall
                     (if (<= (carausius-morosus-tcl2-angle bo) *AEP-L3*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor l3))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-forward-angle-sensor R1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor R1))
                    (funcall
                     (if (<= (carausius-morosus-tcr0-angle bo) *AEP-R1*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor r1))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-forward-angle-sensor R2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor R2))
                    (funcall
                     (if (<= (carausius-morosus-tcr1-angle bo) *AEP-R2*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor r2))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-forward-angle-sensor R3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-forward-angle-sensor R3))
                    (funcall
                     (if (<= (carausius-morosus-tcr2-angle bo) *AEP-R3*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-forward-angle-sensor r3))))))
    ;; Connect backward angle sensors
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor l1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor l1))
                    (funcall
                     (if (>= (carausius-morosus-tcl0-angle bo) *PEP-L1*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor l1))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor l2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor l2))
                    (funcall
                     (if (>= (carausius-morosus-tcl1-angle bo) *PEP-L2*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor l2))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor l3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor l3))
                    (funcall
                     (if (>= (carausius-morosus-tcl2-angle bo) *PEP-L3*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor l3))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor r1))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor r1))
                    (funcall 
                     (if (>= (carausius-morosus-tcr0-angle bo) *PEP-R1*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor r1))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor r2))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor r2))
                    (funcall
                     (if (>= (carausius-morosus-tcr1-angle bo) *PEP-R2*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor r2))))))
    (setf (sensor-neuron-sensor-function
           (leg-controller-backward-angle-sensor r3))
          #'(lambda ()
              (setf (neuron-external-current
                     (leg-controller-backward-angle-sensor r3))
                    (funcall
                     (if (>= (carausius-morosus-tcr2-angle bo) *PEP-R3*) #'+ #'-)
                     (neuron-external-current-magnitude
                      (leg-controller-backward-angle-sensor r3))))))
    ;; Connect stance neurons
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron l1))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl0-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron l2))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl1-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron l3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl2-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron r1))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr0-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron r2))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr1-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-stance-neuron r3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr2-stance bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    ;; Connect swing neurons
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron l1))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl0-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron l2))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl1-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron l3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcl2-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron r1))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr0-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron r2))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr1-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    (setf (motor-neuron-motor-function (leg-controller-swing-neuron r3))
          #'(lambda (firing-frequency)
              (abs (setf (carausius-morosus-tcr2-swing bo)
                         (* firing-frequency *MAXIMUM-LEG-FORCE*)))))
    ))

(defmethod ctrnn::neural-network-neurons ((cns central-nervous-system))
  "List all neurons in system once."
  (remove-duplicates
   (append (neural-network-neurons (central-nervous-system-locomotion-controller cns))
	   (neural-network-neurons (central-nervous-system-lobe cns))))) 