;;;; This file contains class and method definitions for on leg of the carausius-morosus

(in-package :carausius-morosus)

(defclass leg-controller (neural-network)
  ((backward-angle-sensor
    :accessor leg-controller-backward-angle-sensor
    :documentation "Fires continously when leg is in PEP."
    :initarg :backward-angle-sensor
    :initform (make-instance 'sensor-neuron))
   (coxa-angle-sensor
    :accessor leg-controller-coxa-angle-sensor
    :documentation "Fires proportionally to coxa angle."
    :initarg :coxa-angle-sensor
    :initform (make-instance 'sensor-neuron))
   (femur-tibia-neuron
    :accessor leg-controller-femur-tibia-neuron
    :documentation "Motor neuron to stretch/retract femur-tibia joint."
    :initarg :femur-tibia-neuron
    :initform (make-instance 'motor-neuron))
   (foot-neuron
    :accessor leg-controller-foot-neuron
    :documentation "A motor neuron lifting or lowering the leg."
    :initarg :foot-neuron
    :initform (make-instance 'motor-neuron))
   (forward-angle-sensor
    :accessor leg-controller-forward-angle-sensor
    :documentation "Fires continously when the leg is in it's AEP."
    :initarg :forward-angle-sensor
    :initform (make-instance 'sensor-neuron))
   (lc-neuron
    :accessor leg-controller-lc-neuron
    :documentation "Influences the pace of this leg (is shared among legs)."
    :initarg :lc-neuron
    :initform (make-instance 'neuron))
   (pacemaker
    :accessor leg-controller-pacemaker
    :documentation "A neural network that can create rythmic bursts."
    :initarg :pacemaker
    :initform (make-instance 'pacemaker))
   (stance-neuron
    :accessor leg-controller-stance-neuron
    :documentation "Motor neuron that applies force to the stance muscle."
    :initarg :stance-neuron
    :initform (make-instance 'motor-neuron))
   (swing-neuron
    :accessor leg-controller-swing-neuron
    :documentation "A motor neuron that applies force to the swing muscle."
    :initarg :swing-neuron
    :initform (make-instance 'motor-neuron))))

(defmethod initialize-instance :after ((leg-controller leg-controller) &rest args)
  "Connects all neurons together after initialization."
  (declare (ignore args))
;;;;   (add-synapse!
;;;;    (leg-controller-stance-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-lc-neuron leg-controller)
;;;;                   :strength 5))
;;;;   (add-synapse!
;;;;    (pacemaker-io-neuron (leg-controller-pacemaker leg-controller))
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-lc-neuron leg-controller)
;;;;                   :strength 10))
;;;;   (add-synapse!
;;;;    (leg-controller-stance-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (pacemaker-io-neuron
;;;;                                 (leg-controller-pacemaker leg-controller))
;;;;                   :strength -10))
;;;;   (add-synapse!
;;;;    (leg-controller-swing-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (pacemaker-io-neuron
;;;;                                 (leg-controller-pacemaker leg-controller))
;;;;                   :strength 10))
;;;;   (add-synapse!
;;;;    (leg-controller-foot-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (pacemaker-io-neuron
;;;;                                 (leg-controller-pacemaker leg-controller))
;;;;                   :strength -4))
;;;;   (add-synapse!
;;;;    (pacemaker-io-neuron (leg-controller-pacemaker leg-controller))
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-backward-angle-sensor
;;;;                                 leg-controller)
;;;;                   :strength 10))
;;;;   (add-synapse!
;;;;    (pacemaker-io-neuron (leg-controller-pacemaker leg-controller))
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-forward-angle-sensor
;;;;                                 leg-controller)
;;;;                   :strength -15))
;;;;   (add-synapse!
;;;;    (leg-controller-stance-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-forward-angle-sensor
;;;;                                 leg-controller)
;;;;                   :strength 10))
;;;;   (add-synapse!
;;;;    (leg-controller-swing-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-forward-angle-sensor
;;;;                                 leg-controller)
;;;;                   :strength -15))
;;;;   (add-synapse!
;;;;    (leg-controller-foot-neuron leg-controller)
;;;;    (make-instance 'synapse
;;;;                   :from-neuron (leg-controller-forward-angle-sensor
;;;;                                 leg-controller)
;;;;                   :strength 10)))
  )

(defmethod ctrnn:neural-network-neurons ((leg-controller leg-controller))
  "An override of the superclass method, to properly list all neurons."
  (remove-duplicates ; Some neurons may be shared (= appearing several times)
   (append (neural-network-neurons (leg-controller-pacemaker leg-controller))
           (list (leg-controller-backward-angle-sensor leg-controller)
                 (leg-controller-coxa-angle-sensor leg-controller)
                 (leg-controller-femur-tibia-neuron leg-controller)
                 (leg-controller-foot-neuron leg-controller)
                 (leg-controller-forward-angle-sensor  leg-controller)
; NOTE: listed by locomotion controller instead                 (leg-controller-lc-neuron leg-controller)
                 (leg-controller-stance-neuron leg-controller)
                 (leg-controller-swing-neuron  leg-controller)
		 ))))
   
