;;;; What to evolve:
;;;;
;;;; 34 35 36 37:
;;;; - Bias and t.const for right and left touch sensor neurons
;;;; - Bias and t.const for right and left touch "hidden" neuron
;;;;
;;;; 38:
;;;; - Weights from touch sensor neurons to hidden touch neurons
;;;; 
;;;; 39:
;;;; - Weights from hidden touch neurons neurons to output neurons
;;;;
;;;; 40:
;;;; - Weights from front leg angle sensors to hidden touch neurons
;;;;
;;;; 41:
;;;; - Self-weight for hidden neuron
;;;;
;;;; This GA corresponds to the creature in carausius-morosus 8
;;;;

(in-package :navigation)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))

(defun make-random-population (size)
  (loop for n from 1 to size
     collect (let ((individual
		    (loop for gene from 0 to 41
			  ;; Merge initialization settings from pacemaker initialization, leg and turning below:
			  collect
			  (cond 
			    ;; Insert all special cases for all genes, to ensure bifurcation and
			    ;; center-crossing networks
			    ((or (= gene 4) (= gene 6)) ; Self-weight larger than 4 for pacemaker neurons
			     (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
			    ((= gene 8)
			     8388608) ; stance neuron bias
			    ((= n 10)
			     9437184) ; swing neuron bias
			    ((= gene 12)
			     (+ 7600000
				(- 100000
				   (make-random-number-between
				    0 200000)))) ; foot neuron bias
			    ((= gene 14)
			     8388608) ; sensory neurons bias
			    ((= gene 16) ; pacemaker->stance ~-10
			     (let ((middle (/ (expt 2 24) 4)))
			       (make-random-number-between
				(round (- middle 1000000))
				(round (+ middle 1000000)))))
			    ((or (= gene 17) ; pacemaker->swing ~10
				 (= gene 19) ; bw.ang.->pacemaker ~10
				 (= gene 21) ; fw.ang.->stance ~10
				 (= gene 23)) ; fw.ang.->foot ~10
			     (let ((middle (- (expt 2 24)
					      (/ (expt 2 24) 4))))
			       (make-random-number-between
				(round (- middle 1000000))
				(round (+ middle 1000000)))))
			    ((= gene 18) ; pacemaker->foot ~4
			     (let ((quart (/ (expt 2 24) 20/8)))
			       (make-random-number-between
				(round (- quart 1000000))
				(round (+ quart 1000000)))))
			    ((or (= gene 20) ; fw.ang.->pacemaker ~15
				 (= gene 22) ; fw.ang.->swing ~15
				 (= gene 24)) ; pacemakers ~15
			     (let ((quart (/ (expt 2 24) 8)))
			       (make-random-number-between
				(round (- quart 1000000))
				(round (+ quart 1000000)))))
			    ((= gene 31) ; Self weights of hidden neurons > 4
			     (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
			    ((= gene 33) ; Negative connection to stance
			     (random (/ (expt 2 24) 2)))
			    ((= gene 41) ; Self weight > 4
			     (make-random-number-between 
			      (weight->weight-gene 4) (expt 2 24)))
			    (t
			     (random (expt 2 24)))))))
	       ;; Earlier adjustments
	       ;;  Adjust biases and time constants
	       ;; * Pacemaker evolution *
	       (setf (nth 1 individual) (make-random-number-between 450000 600000))
	       (setf (nth 3 individual) (make-random-number-between 450000 600000))
	       (setf (nth 0 individual)
		     (weight->weight-gene
		      (- (/ (+ (weight-gene->weight (nth 4 individual))
			       (weight-gene->weight (nth 5 individual)))
			    2))))
	       (setf (nth 2 individual)
		     (weight->weight-gene
		      (- (/ (+ (weight-gene->weight (nth 6 individual))
			       (weight-gene->weight (nth 7 individual)))
			    2))))
	       ;; Readjust for Obstacle evolving
	       (setf (nth 27 individual)
		     (weight->weight-gene
		      (-
		       (/ 
			(+
			 (weight-gene->weight (nth 29 individual)) ; odour s.1
			 (weight-gene->weight (nth 30 individual)) ; odour s.2
			 (weight-gene->weight (nth 31 individual)) ; self weight
			 (weight-gene->weight (nth 32 individual)) ; other outpt
			 (weight-gene->weight (nth 39 individual))) ; h. touch
			2))))
               ;; Adjust bias for hidden touch neurons
	       (setf (nth 36 individual)
		     (weight->weight-gene
		      (- 
		       (/
			(+
			 (weight-gene->weight (nth 38 individual)) ; touch s.
			 (weight-gene->weight (nth 40 individual)) ;  f. leg s.
			 (weight-gene->weight (nth 41 individual))) ; self wgt.
			2))))
	       (cons (coerce individual 'vector) 0)))) ; Neither fit nor unfit 

(defun genotype->phenotype (genome)
  (let ((cns ; Realize locomotion controller from earlier evolved genome
	 (leg-controller::genotype->phenotype 
	  (cons (subseq genome 0 25) 0)))
	;; Add lobe
	(lobe (make-instance 'neural-network)))
    ;; Attach lobe to existing neural system
    (setf (carausius-morosus:central-nervous-system-lobe cns) lobe)
    ;; Add neurons to the lobe
    (setf (neural-network-neurons lobe)
	  (list (make-instance 'sensor-neuron)
		(make-instance 'sensor-neuron)
		(make-instance 'neuron) ; left output neuron
		(make-instance 'neuron) ; right output neuron
		(make-instance 'sensor-neuron) ; left touch neuron
		(make-instance 'sensor-neuron) ; right touch neuron
		(make-instance 'neuron) ; left hidden touch neuron
		(make-instance 'neuron))) ; right hidden touch neuron
    ;;;; ******* From turning evolution
    ;; Adjust properties of the neurons
    (setf 
     ;; Gene 25
     (neuron-bias (first (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     (neuron-bias (second (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     ;; Gene 26
     (neuron-time-constant (first (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     (neuron-time-constant (second (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     ;; Gene 27
     (neuron-bias (third (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     (neuron-bias (fourth (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     ;; Gene 28
     (neuron-time-constant (third (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28))
     (neuron-time-constant (fourth (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28)))
    ;; Add synapses to the output neurons (gene 29 30 31 32)
    (setf (neuron-dendrites (third (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    (setf (neuron-dendrites (fourth (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe)) ; NOTE: _second_
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe)) ; NOTE: then _first_
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    ;; Add synapses to the middle leg stance neurons (gene 33)
    (push (make-instance 'synapse
			 :from-neuron (third (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites 
	   (leg-controller-stance-neuron 
	    (locomotion-controller-l2
	     (central-nervous-system-locomotion-controller cns)))))
    (push (make-instance 'synapse
			 :from-neuron (fourth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites
	   (leg-controller-stance-neuron
	    (locomotion-controller-r2
	     (central-nervous-system-locomotion-controller cns)))))
    ;; Add sensor function to sensor neurons
    (setf (sensor-neuron-sensor-function (first (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (first (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-fol-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (first (neural-network-neurons lobe))))))
	  (sensor-neuron-sensor-function (second (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (second (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-for-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (second (neural-network-neurons lobe)))))))
    ;;;; ******* Obstacle evolution
    ;; Bias of touch sensor neurons
    (setf (neuron-bias (fifth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 34))
	  (neuron-bias (sixth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 34)))
    ;; Time constant of sensor neurons
    (setf (neuron-time-constant (fifth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 35))
	  (neuron-time-constant (sixth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 35)))
    ;; Bias of hidden touch neurons
    (setf (neuron-bias (seventh (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 36))
	  (neuron-bias (eighth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 36)))
    ;; Time constant of hidden touch neurons
    (setf (neuron-time-constant (seventh (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 37))
	  (neuron-time-constant (eighth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 37)))
    ;; Weights from touch sensor neurons to hidden touch neurons
    (push (make-instance 'synapse
			 :from-neuron (fifth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 38)))
	  (neuron-dendrites (seventh (neural-network-neurons lobe))))
    (push (make-instance 'synapse
			 :from-neuron (sixth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 38)))
	  (neuron-dendrites (eighth (neural-network-neurons lobe))))
    ;; Weights from hidden touch neurons to output neurons
    (push (make-instance 'synapse
			 :from-neuron (seventh (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 39)))
	  (neuron-dendrites (third (neural-network-neurons lobe))))
    (push (make-instance 'synapse
			 :from-neuron (eighth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 39)))
	  (neuron-dendrites (fourth (neural-network-neurons lobe))))
    ;; Weights from front leg angle sensors to hidden touch neurons
    (push (make-instance
	   'synapse
	   :from-neuron (leg-controller-forward-angle-sensor
			 (locomotion-controller-l1
			  (central-nervous-system-locomotion-controller cns)))
			 :strength (weight-gene->weight (svref genome 40)))
	  (neuron-dendrites 
	   (seventh (neural-network-neurons lobe))))
    (push (make-instance
	   'synapse
	   :from-neuron (leg-controller-forward-angle-sensor
			 (locomotion-controller-r1
			  (central-nervous-system-locomotion-controller cns)))
	   :strength (weight-gene->weight (svref genome 40)))
	  (neuron-dendrites
	   (eighth (neural-network-neurons lobe))))
    ;; Self-weight of hidden touch neuron
    (push (make-instance
	   'synapse
	   :from-neuron (seventh (neural-network-neurons lobe))
	   :strength (weight-gene->weight (svref genome 41)))
	  (neuron-dendrites (seventh (neural-network-neurons lobe))))
    (push (make-instance 
	   'synapse
	   :from-neuron (eighth (neural-network-neurons lobe))
	   :strength (weight-gene->weight (svref genome 41)))
	  (neuron-dendrites (eighth (neural-network-neurons lobe))))
    ;; Connect sensor functions of touch sensors
    (setf (sensor-neuron-sensor-function
	   (fifth (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf (neuron-external-current
		     (fifth (neural-network-neurons lobe)))
		    (* (neuron-external-current-magnitude
			(fifth (neural-network-neurons lobe)))
		       (carausius-morosus::carausius-morosus-ftl-strength
			(central-nervous-system-breve-object cns))))))
     (setf (sensor-neuron-sensor-function
	    (sixth (neural-network-neurons lobe)))
	   #'(lambda ()
	       (setf (neuron-external-current
		      (sixth (neural-network-neurons lobe)))
		     (* (neuron-external-current-magnitude
			 (sixth (neural-network-neurons lobe)))
			(carausius-morosus::carausius-morosus-ftr-strength
			 (central-nervous-system-breve-object cns))))))
     cns))

(defun get-distance-to-target ()
  (read-from-string
   (string-trim
    '(\")
    (execute-breve-command-with-result "get-distance-to-target"))))

(defun get-time ()
  (read-from-string
   (string-trim 
    '(\") (breve-interface:execute-breve-command-with-result "get-time"))))

(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun fitness (genotype &optional (num-breve-synchs 200))
  (unless (gethash (car genotype) *fitnesses*) ; Eval. only if not already ev. 
    (setf
     (gethash (car genotype) *fitnesses*)
     (progn
;       (breve-interface:execute-breve-command "reset")
       (breve-interface:execute-breve-command "set-output-filter" -1)
       (breve-interface:wait-for-breve) ; Wait for breve to finish calculations
       (if (= (random 2) 1) ; 50% chance for testing either right or left
	   (progn
	     (execute-breve-command "reset")
	     (execute-breve-command "move-wall-left")
	     (execute-breve-command "move-odour-patch-left"))
	   (progn
	     (execute-breve-command "reset")
	     (wait-for-breve)
	     (execute-breve-command "move-wall-right")
	     (execute-breve-command "move-odour-patch-right")))
       (breve-interface:wait-for-breve)
       (let ((energy 0)
	     (cns (genotype->phenotype (car genotype)))
	     (start-distance (get-distance-to-target)))
	 (dotimes (var num-breve-synchs)
	   (dotimes (var 5)
	     (incf energy
		   (synchronously-update-membrane-potentials!
	       (central-nervous-system-locomotion-controller cns)))
	     (incf energy
		   (synchronously-update-membrane-potentials!
		    (central-nervous-system-lobe cns))))
	   (synchronize-with-breve (central-nervous-system-breve-object cns))
	   (breve-interface:wait-for-breve))
	 ;; Return improvement in position
	 (let ((new-distance (get-distance-to-target)))
	   (- start-distance new-distance))))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
     and num from 1
     collect (progn
	       (format t "Evaluating ~A (~A of ~A):~% "
		       (car genotype) num (length individuals))
	       (let ((res (cons (car genotype)
				(fitness genotype num-breve-synchs))))
		 (format t " ~A~%" (cdr res))
		 res))))
		
(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent (nth (make-random-number-between
				      0 (- (round (/ (length population) 4))
					   parent-num)) population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
    (let* ((carausius-morosus::*maximum-leg-force* 200000)
	   (carausius-morosus::*maximum-leg-down-force* 80000)
	   (population 
	    (sort-population (evaluate (make-random-population 100) 500))))
      (with-open-file (f (format nil "logs/obstacle-without-increment-log~A.txt" id) 
			 :direction :output :if-exists :supersede)
	(do ((generation 1 (+ generation 1)))
	    ((> generation num-generations))
	  (format t "*** Generation ~A~%" generation)
	  ;; Select parents
	  (let ((parents (select-parents population 2)))
	    ;; Recombine
	    (let ((children (recombine parents)))
	      ;; Maybe mutate
	      (when (< (random 100) 5)
		(setf children (loop for c in children collect (mutate c))))
	      ;; Evaluate new individuals
	      (setf children (evaluate children 500))
	      ;; Select individuals for next generation
	      (setf population
		    (sort-population (append
				      children
				      (remove-individuals population 2))))
	      ;; Update display
	      (format f "~& ~A ~A ~A ~A ~A~%"
		      generation
		      (cdr (first population)) ; Fitness of best
		      (/ (loop for ind in population
			    sum (cdr ind))
			 (length population))
		      (cdr (first (last population)))
		      (car (first population))) ;; Genome of best
	      (force-output f)))))
      (first population)))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (dotimes (var 10)
    (when (> var 7)
      (ga 900 var))))
