;;;; What to evolve:
;;;;
;;;; 34 35 36 37:
;;;; - Bias and t.const for right and left touch sensor neurons
;;;; - Bias and t.const for right and left touch "hidden" neuron
;;;;
;;;; 38:
;;;; - Weights from touch sensor neurons to hidden touch neurons
;;;; 
;;;; 39:
;;;; - Weights from hidden touch neurons neurons to output neurons
;;;;
;;;; 40:
;;;; - Weights from front leg angle sensors to hidden touch neurons
;;;;
;;;; 41:
;;;; - Self-weight for hidden neuron
;;;;
;;;; This GA corresponds to the creature in carausius-morosus 8
;;;;

(in-package :navigation)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))

(defun make-random-population (size)
  (loop for n from 1 to size
     collect (let ((individual
		    (loop for gene from 0 to 41
		       collect (cond 
				 ((= gene 41) ; Self weight > 4
				  (make-random-number-between 
				   (weight->weight-gene 4) (expt 2 24)))
				 (t
				  (random (expt 2 24)))))))
	       ;; Inject genome from turningGA
	       (setf 
                (subseq individual 0 34)
                (list
                 10971944 521977 2457366 482522 11388430 254725 12565966
                 16020378 8418631 16706274 9429669 1443941 7522381 501240 
                 8398402 5077330 3797499 12301242 6525757 13024086 2317410
                 13145069 1489847 12814818 1233111 9887649 3083939 7701638 
                 1319205 14399656 587349 10801335 6689238 533015))
;		 6707912 519146 5847706 475293 11021734 4034557 13209047
;		 5113294 8388608 1413607 10518733 1691558 7664068 1255015
;		 8388608 2409152 5188667 12542669 6471289 12766363 2653418
;		 13374513 2587161 12544360 1287105 10276338 13811141 2285084 
;		 538246 14533227 10895421 11085297 1142809 4815469))
	       ;; Randomize in neighbourhood of earlier evolved genome
	       (loop for gene-nr from 0 to 33
		  do (incf (nth gene-nr individual)
			   (- 100000 (random 200000))))
	       ;; Ensure individual represent a center crossing net by
	       ;; (re)adjusting bias of outputs to (- (/ sum-of-weights 2))
	       (setf (nth 27 individual)
		     (weight->weight-gene
		      (-
		       (/ 
			(+
			 (weight-gene->weight (nth 29 individual)) ; odour s.1
			 (weight-gene->weight (nth 30 individual)) ; odour s.2
			 (weight-gene->weight (nth 31 individual)) ; self weight
			 (weight-gene->weight (nth 32 individual)) ; other outpt
			 (weight-gene->weight (nth 39 individual))) ; h. touch
			2))))
               ;; Adjust bias for hidden touch neurons
	       (setf (nth 36 individual)
		     (weight->weight-gene
		      (- 
		       (/
			(+
			 (weight-gene->weight (nth 38 individual)) ; touch s.
			 (weight-gene->weight (nth 40 individual)) ;  f. leg s.
			 (weight-gene->weight (nth 41 individual))) ; self wgt.
			2))))
	       (cons (coerce individual 'vector) 0)))) ; Neither fit nor unfit 

(defun genotype->phenotype (genome)
  (let ((cns ; Realize locomotion controller from earlier evolved genome
	 (leg-controller::genotype->phenotype 
	  (cons (subseq genome 0 25) 0)))
	;; Add lobe
	(lobe (make-instance 'neural-network)))
    ;; Attach lobe to existing neural system
    (setf (carausius-morosus:central-nervous-system-lobe cns) lobe)
    ;; Add neurons to the lobe
    (setf (neural-network-neurons lobe)
	  (list (make-instance 'sensor-neuron)
		(make-instance 'sensor-neuron)
		(make-instance 'neuron) ; left output neuron
		(make-instance 'neuron) ; right output neuron
		(make-instance 'sensor-neuron) ; left touch neuron
		(make-instance 'sensor-neuron) ; right touch neuron
		(make-instance 'neuron) ; left hidden touch neuron
		(make-instance 'neuron))) ; right hidden touch neuron
    ;;;; ******* From turning evolution
    ;; Adjust properties of the neurons
    (setf 
     ;; Gene 25
     (neuron-bias (first (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     (neuron-bias (second (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     ;; Gene 26
     (neuron-time-constant (first (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     (neuron-time-constant (second (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     ;; Gene 27
     (neuron-bias (third (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     (neuron-bias (fourth (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     ;; Gene 28
     (neuron-time-constant (third (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28))
     (neuron-time-constant (fourth (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28)))
    ;; Add synapses to the output neurons (gene 29 30 31 32)
    (setf (neuron-dendrites (third (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    (setf (neuron-dendrites (fourth (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe)) ; NOTE: _second_
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe)) ; NOTE: then _first_
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    ;; Add synapses to the middle leg stance neurons (gene 33)
    (push (make-instance 'synapse
			 :from-neuron (third (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites 
	   (leg-controller-stance-neuron 
	    (locomotion-controller-l2
	     (central-nervous-system-locomotion-controller cns)))))
    (push (make-instance 'synapse
			 :from-neuron (fourth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites
	   (leg-controller-stance-neuron
	    (locomotion-controller-r2
	     (central-nervous-system-locomotion-controller cns)))))
    ;; Add sensor function to sensor neurons
    (setf (sensor-neuron-sensor-function (first (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (first (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-fol-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (first (neural-network-neurons lobe))))))
	  (sensor-neuron-sensor-function (second (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (second (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-for-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (second (neural-network-neurons lobe)))))))
    ;;;; ******* Obstacle evolution
    ;; Bias of touch sensor neurons
    (setf (neuron-bias (fifth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 34))
	  (neuron-bias (sixth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 34)))
    ;; Time constant of sensor neurons
    (setf (neuron-time-constant (fifth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 35))
	  (neuron-time-constant (sixth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 35)))
    ;; Bias of hidden touch neurons
    (setf (neuron-bias (seventh (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 36))
	  (neuron-bias (eighth (neural-network-neurons lobe)))
	  (weight-gene->weight (svref genome 36)))
    ;; Time constant of hidden touch neurons
    (setf (neuron-time-constant (seventh (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 37))
	  (neuron-time-constant (eighth (neural-network-neurons lobe)))
	  (time-constant-gene->time-constant (svref genome 37)))
    ;; Weights from touch sensor neurons to hidden touch neurons
    (push (make-instance 'synapse
			 :from-neuron (fifth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 38)))
	  (neuron-dendrites (seventh (neural-network-neurons lobe))))
    (push (make-instance 'synapse
			 :from-neuron (sixth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 38)))
	  (neuron-dendrites (eighth (neural-network-neurons lobe))))
    ;; Weights from hidden touch neurons to output neurons
    (push (make-instance 'synapse
			 :from-neuron (seventh (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 39)))
	  (neuron-dendrites (third (neural-network-neurons lobe))))
    (push (make-instance 'synapse
			 :from-neuron (eighth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 39)))
	  (neuron-dendrites (fourth (neural-network-neurons lobe))))
    ;; Weights from front leg angle sensors to hidden touch neurons
    (push (make-instance
	   'synapse
	   :from-neuron (leg-controller-forward-angle-sensor
			 (locomotion-controller-l1
			  (central-nervous-system-locomotion-controller cns)))
			 :strength (weight-gene->weight (svref genome 40)))
	  (neuron-dendrites 
	   (seventh (neural-network-neurons lobe))))
    (push (make-instance
	   'synapse
	   :from-neuron (leg-controller-forward-angle-sensor
			 (locomotion-controller-r1
			  (central-nervous-system-locomotion-controller cns)))
	   :strength (weight-gene->weight (svref genome 40)))
	  (neuron-dendrites
	   (eighth (neural-network-neurons lobe))))
    ;; Self-weight of hidden touch neuron
    (push (make-instance
	   'synapse
	   :from-neuron (seventh (neural-network-neurons lobe))
	   :strength (weight-gene->weight (svref genome 41)))
	  (neuron-dendrites (seventh (neural-network-neurons lobe))))
    (push (make-instance 
	   'synapse
	   :from-neuron (eighth (neural-network-neurons lobe))
	   :strength (weight-gene->weight (svref genome 41)))
	  (neuron-dendrites (eighth (neural-network-neurons lobe))))
    ;; Connect sensor functions of touch sensors
    (setf (sensor-neuron-sensor-function
	   (fifth (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf (neuron-external-current
		     (fifth (neural-network-neurons lobe)))
		    (* (neuron-external-current-magnitude
			(fifth (neural-network-neurons lobe)))
		       (carausius-morosus::carausius-morosus-ftl-strength
			(central-nervous-system-breve-object cns))))))
     (setf (sensor-neuron-sensor-function
	    (sixth (neural-network-neurons lobe)))
	   #'(lambda ()
	       (setf (neuron-external-current
		      (sixth (neural-network-neurons lobe)))
		     (* (neuron-external-current-magnitude
			 (sixth (neural-network-neurons lobe)))
			(carausius-morosus::carausius-morosus-ftr-strength
			 (central-nervous-system-breve-object cns))))))
     cns))

(defun get-distance-to-target ()
  (read-from-string
   (string-trim
    '(\")
    (execute-breve-command-with-result "get-distance-to-target"))))

(defun get-time ()
  (read-from-string
   (string-trim 
    '(\") (breve-interface:execute-breve-command-with-result "get-time"))))

(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun fitness (genotype &optional (num-breve-synchs 200))
  (unless (gethash (car genotype) *fitnesses*) ; Eval. only if not already ev. 
    (setf
     (gethash (car genotype) *fitnesses*)
     (progn
;       (breve-interface:execute-breve-command "reset")
       (breve-interface:execute-breve-command "set-output-filter" -1)
       (breve-interface:wait-for-breve) ; Wait for breve to finish calculations
       (if (= (random 2) 1) ; 50% chance for testing either right or left
	   (progn
	     (execute-breve-command "reset")
	     (execute-breve-command "move-wall-left")
	     (execute-breve-command "move-odour-patch-left"))
	   (progn
	     (execute-breve-command "reset")
	     (wait-for-breve)
	     (execute-breve-command "move-wall-right")
	     (execute-breve-command "move-odour-patch-right")))
       (breve-interface:wait-for-breve)
       (let ((energy 0)
	     (cns (genotype->phenotype (car genotype)))
	     (start-distance (get-distance-to-target)))
	 (dotimes (var num-breve-synchs)
	   (dotimes (var 5)
	     (incf energy
		   (synchronously-update-membrane-potentials!
	       (central-nervous-system-locomotion-controller cns)))
	     (incf energy
		   (synchronously-update-membrane-potentials!
		    (central-nervous-system-lobe cns))))
	   (synchronize-with-breve (central-nervous-system-breve-object cns))
	   (breve-interface:wait-for-breve))
	 ;; Return improvement in position
	 (let ((new-distance (get-distance-to-target)))
	   (- start-distance new-distance))))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
     and num from 1
     collect (progn
	       (format t "Evaluating ~A (~A of ~A):~% "
		       (car genotype) num (length individuals))
	       (let ((res (cons (car genotype)
				(fitness genotype num-breve-synchs))))
		 (format t " ~A~%" (cdr res))
		 res))))
		
(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent (nth (make-random-number-between
				      0 (- (round (/ (length population) 4))
					   parent-num)) population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
    (let* ((carausius-morosus::*maximum-leg-force* 200000)
	   (carausius-morosus::*maximum-leg-down-force* 80000)
	   (population 
	    (sort-population (evaluate (make-random-population 100) 500))))
      (with-open-file (f (format nil "logs/obstacle-log~A.txt" id) 
			 :direction :output :if-exists :supersede)
	(do ((generation 1 (+ generation 1)))
	    ((> generation num-generations))
	  (format t "*** Generation ~A~%" generation)
	  ;; Select parents
	  (let ((parents (select-parents population 2)))
	    ;; Recombine
	    (let ((children (recombine parents)))
	      ;; Maybe mutate
	      (when (< (random 100) 5)
		(setf children (loop for c in children collect (mutate c))))
	      ;; Evaluate new individuals
	      (setf children (evaluate children 500))
	      ;; Select individuals for next generation
	      (setf population
		    (sort-population (append
				      children
				      (remove-individuals population 2))))
	      ;; Update display
	      (format f "~& ~A ~A ~A ~A ~A~%"
		      generation
		      (cdr (first population)) ; Fitness of best
		      (/ (loop for ind in population
			    sum (cdr ind))
			 (length population))
		      (cdr (first (last population)))
		      (car (first population))) ;; Genome of best
	      (force-output f)))))
      (first population)))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (dotimes (var 10)
    (when (= var 7);  (and (< var 7) (> var 5))
      (ga 250 var))))

(defun test-run ()
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (fitness
     (cons
      ;; (vector 3733533 579450 9756342 537173 11260656 14728187 10562126 3479623 8388608 11265309 9362421 16651737 7651600 2495017 8388608 1321523 3804438 13513611 6565785 11953726 2243433 13514748 2165012 12145827 1781507 8974017 371559 14758130 14770825 6430388 1841693 14305577 5332363 8304389 6080989 3014895 -943646 12527613 10915402 3534011 7609923 13732556) ; Log 0
      ;; (vector 6329032 505357 8849552 507250 15359226 5575945 10268908 4170133 8388608 12341328 10859663 5455551 7632490 3865638 8388608 1365172 3964924 13348321 6501813 12164745 2648236 12870507 1165343 12242532 1734761 7849078 5733275 4570034 3463469 14654027 -44163 12942644 4383660 4972342 15587297 12139168 3510575 3638984 16241672 11197369 10148713 11422214)
      ;; (vector 2494518 531349 3020269 625776 15928182 12699369 14727086 6062998 8388608 2828890 10714674 7230441 7582175 1261459 8388608 3205106 3528778 13376827 5855691 12425608 1467392 11588646 1761148 12178129 2995291 11434139 1838501 6203 2484660 968101 2065893 16380502 2404936 3547456 9977916 15554754 2284664 1238345 16528196 276197 7849324 11883859)
      ;; (vector 2163608 585115 6594039 553743 15340997 13886219 15623662 4742692 8388608 10985423 11195325 1207518 7640345 1842768 8388608 3281762 3930926 12741482 7575045 13506475 1763073 12570230 2766324 13464829 1763332 12227382 5827259 4505908 1621073 4945762 5105735 14855514 16198166 2435540 4147083 13427234 4618238 15633502 13920220 5930873 7171403 11546543)
      ;; (vector 1855026 595511 8271656 476328 13601990 1661715 13668559 7737033 8388608 2237115 10110645 7067249 7552699 1471115 8388608 2100585 4261666 13110931 6585054 13122052 2541672 12665145 1172468 12895804 1891223 8460536 15436531 -1130238 4545947 13813939 14839663 12360526 1277325 6916161 10460760 15216817 6905568 13640667 3214850 9991372 10023597 16482742)
      ;; (vector 9106708 497346 1880984 471412 10594990 13984917 13955099 12867154 8388608 4398206 7956057 7040754 7694518 333250 8388608 6218283 3527593 13542020 5903410 11769342 1201270 13360739 2938315 12314884 1665082 14111133 15095970 3390858 9156888 12231527 6473074 15780499 381209 7460178 7769910 6201721 11603707 7517403 271608 6905457 3000833 15212575) ;; Log 6
      ;; (vector 7652104 455586 3890890 564405 12770051 5480174 14907040 10875722 8388608 13827552 9217937 5329373 7551099 1421918 8388608 10105247 4208319 12269348 5774065 13166744 2530905 12812310 1586802 12665732 2312597 16696050 8586244 8342442 9107775 5333979 8642751 11486775 14152001 4115371 8571982 4829332 8904872 14419664 1201802 3065895 9147606 13699347) ;; Log 7
      ;; (vector 6258502 586896 1348947 619343 12732763 8304664 15357957 15498581 8388608 2369844 9116660 1902335 7534074 477328 8388608 11579218 4645978 12394178 6094827 12629289 1389911 12881011 1467733 12206025 2901036 5033123 4022419 7146650 10286716 12832837 2374905 14046910 12922473 1924623 2491680 524983 6968360 13933995 15133814 2249830 6076798 13585710) ;; Log 8
      (vector 3243064 526868 2586546 490891 16363588 2010226 13847844 13202634 8388608 4771199 11142939 3156214 7543276 1728864 8388608 4865687 3358493 11705805 7018941 13476141 2479935 11723955 1272251 12008166 2258184 3459603 10720503 -1798867 1559215 11680142 8381503 13738396 6278100 6852848 11899841 12549875 11001906 7720627 6097540 6704954 1880142 14387884)
      0)
     500)))

(defun test-run-good ()
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (fitness
     (cons
      (vector 10881798 464761 2546047 529695 11485290 346478 12645900 15980601 8462712 16773794 9520890 1503890 7586079 560180 8407916 5126934 3789736 12226889 6464642 13090531 2261127 13230912 1479865 12836107 1257511 9827806 3145997 6157148 1321390 14361529 600876 10782515 6734727 533618 3290243 3936455 348182 13589314 14708290 10117556 7648289 13113815)
      0)
     500)))