;;;; What to evolve: EVERYTHING (in one chunk)

(in-package :navigation)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))

(defun make-random-population (size)
  (loop for n from 1 to size
     collect
     (let
         ((individual 
           (loop for gene from 0 to 33
                 collect
                 (cond 
                  ;; Insert all special cases for all genes, to ensure bifurcation and
                  ;; center-crossing networks
                  ((or (= gene 4) (= gene 6)) ; Self-weight larger than 4 for pacemaker neurons
                   (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
                  ((= gene 8)
                   8388608) ; stance neuron bias
                  ((= n 10)
                   9437184) ; swing neuron bias
                  ((= gene 12)
                   (+ 7600000 
                      (- 100000 
                         (make-random-number-between
                          0 200000)))) ; foot neuron bias
                  ((= gene 14)
                   8388608) ; sensory neurons bias
                  ((= gene 16) ; pacemaker->stance ~-10
                   (let ((middle (/ (expt 2 24) 4)))
                     (make-random-number-between
                      (round (- middle 1000000))
                      (round (+ middle 1000000)))))
                  ((or (= gene 17) ; pacemaker->swing ~10
                       (= gene 19) ; bw.ang.->pacemaker ~10
                       (= gene 21) ; fw.ang.->stance ~10
                       (= gene 23)) ; fw.ang.->foot ~10
                   (let ((middle (- (expt 2 24)
                                    (/ (expt 2 24) 4))))
                     (make-random-number-between
                      (round (- middle 1000000))
                      (round (+ middle 1000000)))))
                  ((= gene 18) ; pacemaker->foot ~4
                   (let ((quart (/ (expt 2 24) 20/8)))
                     (make-random-number-between 
                      (round (- quart 1000000))
                      (round (+ quart 1000000)))))
                  ((or (= gene 20) ; fw.ang.->pacemaker ~15 
                       (= gene 22) ; fw.ang.->swing ~15
                       (= gene 24)) ; pacemakers ~15
                   (let ((quart (/ (expt 2 24) 8)))
                     (make-random-number-between 
                      (round (- quart 1000000))
                      (round (+ quart 1000000)))))
                  ((= gene 31) ; Self weights of hidden neurons > 4
                   (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
                  ((= gene 33) ; Negative connection to stance
                   (random (/ (expt 2 24) 2)))
                  (t
                   (random (expt 2 24)))))))
       ;;;;  Adjust biases and time constants
       ;; * Pacemaker evolution *
       (setf (nth 1 individual) (make-random-number-between 450000 600000))
       (setf (nth 3 individual) (make-random-number-between 450000 600000))
       (setf (nth 0 individual)
             (weight->weight-gene
              (- (/ (+ (weight-gene->weight (nth 4 individual))
                       (weight-gene->weight (nth 5 individual)))
                    2))))
       (setf (nth 2 individual)
             (weight->weight-gene
              (- (/ (+ (weight-gene->weight (nth 6 individual))
                       (weight-gene->weight (nth 7 individual)))
                    2))))
       ;; Ensure individual represent a center crossing net by adjusting bias of outputs to (- (/ sum-of-weights 2))
	       (setf (nth 27 individual)
		     (weight->weight-gene
		      (- (/ (+ (weight-gene->weight (nth 29 individual)) ; from sensor 1
			       (weight-gene->weight (nth 30 individual)) ; from sensor 2
			       (weight-gene->weight (nth 31 individual)) ; self weight
			       (weight-gene->weight (nth 32 individual))) ; from other output
			    2))))
	       (cons (coerce individual 'vector) 0)))) ; Neither fit nor unfit 

(defun genotype->phenotype (genome)
  (let ((cns (leg-controller::genotype->phenotype 
	      ;; Realize locomotion controller from earlier evolved genome
	      (cons (subseq genome 0 25) 0)))
	;; Add a lobe
	(lobe (make-instance 'neural-network)))
    ;; Add lobe to neural system
    (setf (carausius-morosus:central-nervous-system-lobe cns) lobe)
    ;; Add neurons to the lobe
    (setf (neural-network-neurons lobe)
	  (list (make-instance 'sensor-neuron)
		(make-instance 'sensor-neuron)
		(make-instance 'neuron)
		(make-instance 'neuron)))
    ;; Adjust properties of the neurons
    (setf 
     ;; Gene 25
     (neuron-bias (first (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     (neuron-bias (second (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     ;; Gene 26
     (neuron-time-constant (first (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     (neuron-time-constant (second (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     ;; Gene 27
     (neuron-bias (third (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     (neuron-bias (fourth (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     ;; Gene 28
     (neuron-time-constant (third (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28))
     (neuron-time-constant (fourth (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28)))
    ;; Add synapses to the output neurons (gene 29 30 31 32)
    (setf (neuron-dendrites (third (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    (setf (neuron-dendrites (fourth (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe)) ; NOTE: _second_
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe)) ; NOTE: then _first_
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    ;; Add synapses to the middle leg stance neurons (gene 33)
    (push (make-instance 'synapse
			 :from-neuron (third (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites 
	   (leg-controller-stance-neuron 
	    (locomotion-controller-l2
	     (central-nervous-system-locomotion-controller cns)))))
    (push (make-instance 'synapse
			 :from-neuron (fourth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites
	   (leg-controller-stance-neuron
	    (locomotion-controller-r2
	     (central-nervous-system-locomotion-controller cns)))))
    ;; Add sensor function to sensor neurons
    (setf (sensor-neuron-sensor-function (first (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (first (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-fol-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (first (neural-network-neurons lobe))))))
	  (sensor-neuron-sensor-function (second (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (second (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-for-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (second (neural-network-neurons lobe)))))))
    cns))

(defun get-distance-to-target ()
  (read-from-string
   (string-trim
    '(\")
    (execute-breve-command-with-result "get-distance-to-target"))))

(defun get-time ()
  (read-from-string
   (string-trim 
    '(\") (breve-interface:execute-breve-command-with-result "get-time"))))

(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun fitness (genotype &optional (num-breve-synchs 200))
  (unless (gethash (car genotype) *fitnesses*)
    (setf
     (gethash (car genotype) *fitnesses*)
     ;; Go left first, then right
     (let ((fitnesses
	    (loop for var from 0 to 1
		  collect (progn
			    (breve-interface:execute-breve-command "set-output-filter" -1)
			      (execute-breve-command "reset")
			      (if (= var 0)
				  (execute-breve-command "move-odour-patch-left")
				  (execute-breve-command "move-odour-patch-right"))
			      (wait-for-breve)
			      ;; Start evaluation
			      (let ((energy 0)
				    (cns (genotype->phenotype (car genotype)))
				    (start-distance (get-distance-to-target)))
				(dotimes (var num-breve-synchs)
				  (dotimes (var 5)
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-locomotion-controller cns)))
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-lobe cns))))
				  (synchronize-with-breve (central-nervous-system-breve-object cns))
				  (breve-interface:wait-for-breve))
				;; Return improvement in position
				(let ((new-distance (get-distance-to-target)))
				  (- start-distance new-distance)))))))
       (format t "~A~%" fitnesses)
       (apply #'+ fitnesses))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
				   (fitness genotype num-breve-synchs))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
    (let* ((carausius-morosus::*maximum-leg-force* 200000)
	   (carausius-morosus::*maximum-leg-down-force* 80000)
	   (population 
	    (sort-population (evaluate (make-random-population 100) 400))))
      (with-open-file (f (format nil "logs/turning-without-increment-log~A.txt" id) :direction :output :if-exists :supersede)
	(do ((generation 1 (+ generation 1)))
	    ((> generation num-generations))
	  (format t "*** Generation ~A~%" generation)
	  ;; Select parents
	  (let ((parents (select-parents population 2)))
	    ;; Recombine
	    (let ((children (recombine parents)))
	      ;; Maybe mutate
	      (when (< (random 100) 5)
		(setf children (loop for c in children collect (mutate c))))
	      ;; Evaluate new individuals
	      (setf children (evaluate children 400))
	      ;; Select individuals for next generation
	      (setf population
		    (sort-population (append
				      children
				      (remove-individuals population 2))))
	      ;; Update display
	      (format f "~A ~A ~A ~A ~A~%"
		      generation
		      (cdr (first population)) ; Fitness of best
		      (/ (loop for ind in population
			    sum (cdr ind))
			 (length population))
		      (cdr (first (last population)))
		      (car (first population))) ;; Genome of best
	      (force-output f)))))
      (first population)))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (dotimes (var 10)
    (when (> var 7)
      (ga 1000 var))))

(defun test-run (genome)
  (evaluate (list (cons genome 0)) 400))

(defun test-runs ()
  "Test the 10 genome evolved after 650 generations."
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (test-run
     (vector 4053863 499255 3228771 656413 16248772 1499860 16095030 3721210 8388608 11441104 11635203 5131256 7684425 7177354 8388608 4961108 4207603 13107960 5959885 11869540 2496060 11979438 1137073 12252105 2466436 5490987 12543943 10576886 2952866 1567403 14193848 10396452 3225969 6570732))
    (test-run 
     (vector 4020174 474655 4342611 501820 12024751 13489333 13480026 11389184 8388608 5510617 9992114 6701486 7620194 4062264 8388608 858571 3831814 12288332 6570667 11972272 1425826 12436342 2853117 12353725 2090107 1843289 3824978 2829549 9428506 13026160 16490399 11547314 6645661 1477112))
    (test-run
     (vector 8463338 547742 5308152 498628 12147094 5640471 13781228 9156900 8388608 5621834 9630461 9241437 7675465 3368727 8388608 4931896 4210086 11821639 7411686 12824476 2706743 12073519 1730158 11702463 1977680 9068325 2662487 4964825 6943504 13222912 5484764 16207014 5622225 5151162))
    (test-run
     (vector 5110806 450807 3729480 547953 15164629 8168192 10407721 15687750 8388608 13670262 9295115 7017443 7503646 4246318 8388608 557008 3750891 11665666 7249108 12357788 1320758 12101201 2832497 12990937 1526652 10501129 738007 10641918 12878132 13831351 14095742 16771697 6845869 7177892))
    (test-run
     (vector 6910002 544330 4391005 517837 16400388 7724698 10320171 8287726 8388608 14284377 11489204 3602208 7592949 5842237 8388608 1468668 3478312 12615154 7652625 12823573 2762880 12991636 1893362 12410117 2618425 5885480 1841617 2237728 15860236 4555455 790739 13940557 15202134 1398407)
     )
    (test-run
     (vector 10728040 591617 4100629 565044 10995590 1102763 14689927 10741516 8388608 12882636 10578028 7940304 7537417 9645611 8407068 4997960 3222570 12088976 6202357 13274572 1682992 12183267 1548384 13382229 1151703 7402238 11023048 11926690 2344953 11465705 2755112 10138000 2119452 3803599))
    (test-run
     (vector 9033716 490947 2979997 514748 14242789 1244211 12686542 14907896 8388608 11556926 8072953 4545814 7657889 3102873 8388608 888295 3831593 13307408 7185856 12881636 1765969 12729504 1530827 12937753 2633497 6714800 319748 5393276 5272551 16383554 984342 12647483 9610782 4439074))
    (test-run
     (vector 6644112 529146 8647019 471465 10787133 14388112 12168692 4091702 8388608 15195432 9180535 14574676 7575740 2286844 8413170 1169599 5059068 12078223 6226505 12653227 1463325 12574615 1135681 12852522 1429397 7852707 199488 5881858 1174577 13971950 917177 15786468 3100593 2710517))
    (test-run
     (vector 2324017 476611 10420360 486246 12909847 15996551 12292478 421233 8388608 4772000 8433665 10331333 7671056 440956 8388608 1341692 4755672 11740919 7152673 12185948 1240990 12954176 2134647 11682312 2497933 12180398 14015101 5213042 16404183 11030649 16464939 10228546 1705343 4231558))
    ))
