;; Used to test the GA . It writes position of quickbeam every breve synch

(in-package :navigation)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))

(defun make-random-population (size)
  (loop for n from 1 to size
     collect (let ((individual (loop for gene from 0 to 33
				  collect (cond 
					    ((= gene 31) ; Self weights > 4
					     (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
					    ((= gene 33) ; Negative connection to stance
					     (random (/ (expt 2 24) 2)))
					    (t
					     (random (expt 2 24)))))))
	       ;; Inject earlier evolved genome into the population
	       (setf (subseq individual 0 25)
		     (list
		      10963896 516852 2460398 475135 11385571 260912 12572387 16024549 
		      8426675 16714521 9437184 1445525 7527271 493102 8388608 5077806
		      3802889 12303830 6529969 13024977 2308303 13138285 1497664 12821677
		      1233568))
	       ;; Randomize earlier evolved genome to "neighbourhood" by mutation
	       (loop for n from 0 to 24
		     do (incf (nth n individual) (- (random 20000) 10000)))
	       ;; Ensure individual represent a center crossing net by adjusting bias of outputs to (- (/ sum-of-weights 2))
;	       (setf (nth 27 individual)
;		     (weight->weight-gene
;		      (- (/ (+ (weight-gene->weight (nth 29 individual)) ; from sensor 1
;			       (weight-gene->weight (nth 30 individual)) ; from sensor 2
;			       (weight-gene->weight (nth 31 individual)) ; self weight
;			       (weight-gene->weight (nth 32 individual))) ; from other output
;			    2))))
	       (cons (coerce individual 'vector) 0)))) ; Neither fit nor unfit 

(defun genotype->phenotype (genome)
  (let ((cns (leg-controller::genotype->phenotype 
	      ;; Realize locomotion controller from earlier evolved genome
	      (cons (subseq genome 0 25) 0)))
	;; Add a lobe
	(lobe (make-instance 'neural-network)))
    ;; Add lobe to neural system
    (setf (carausius-morosus:central-nervous-system-lobe cns) lobe)
    ;; Add neurons to the lobe
    (setf (neural-network-neurons lobe)
	  (list (make-instance 'sensor-neuron)
		(make-instance 'sensor-neuron)
		(make-instance 'neuron)
		(make-instance 'neuron)))
    ;; Adjust properties of the neurons
    (setf 
     ;; Gene 25
     (neuron-bias (first (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     (neuron-bias (second (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     ;; Gene 26
     (neuron-time-constant (first (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     (neuron-time-constant (second (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     ;; Gene 27
     (neuron-bias (third (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     (neuron-bias (fourth (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     ;; Gene 28
     (neuron-time-constant (third (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28))
     (neuron-time-constant (fourth (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28)))
    ;; Add synapses to the output neurons (gene 29 30 31 32)
    (setf (neuron-dendrites (third (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    (setf (neuron-dendrites (fourth (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe)) ; NOTE: _second_
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe)) ; NOTE: then _first_
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    ;; Add synapses to the middle leg stance neurons (gene 33)
    (push (make-instance 'synapse
			 :from-neuron (third (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites 
	   (leg-controller-stance-neuron 
	    (locomotion-controller-l2
	     (central-nervous-system-locomotion-controller cns)))))
    (push (make-instance 'synapse
			 :from-neuron (fourth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites
	   (leg-controller-stance-neuron
	    (locomotion-controller-r2
	     (central-nervous-system-locomotion-controller cns)))))
    ;; Add sensor function to sensor neurons
    (setf (sensor-neuron-sensor-function (first (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (first (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-fol-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (first (neural-network-neurons lobe))))))
	  (sensor-neuron-sensor-function (second (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (second (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-for-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (second (neural-network-neurons lobe)))))))
    cns))

(defun get-distance-to-target ()
  (read-from-string
   (string-trim
    '(\")
    (execute-breve-command-with-result "get-distance-to-target"))))

(defun get-time ()
  (read-from-string
   (string-trim 
    '(\") (breve-interface:execute-breve-command-with-result "get-time"))))

(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun fitness (genotype &optional (num-breve-synchs 200))
  (unless (gethash (car genotype) *fitnesses*)
    (setf
     (gethash (car genotype) *fitnesses*)
     ;; Go left first, then right
     (let ((fitnesses
	    (loop for var from 0 to 0
		  collect (progn
			    (breve-interface:execute-breve-command "set-output-filter" -1)
			      (execute-breve-command "reset")
			      (if (= var 0)
				  (execute-breve-command "move-odour-patch-left")
				  (execute-breve-command "move-odour-patch-right"))
			      (wait-for-breve)
			      ;; Start evaluation
			      (let ((energy 0)
				    (cns (genotype->phenotype (car genotype)))
				    (start-distance (get-distance-to-target)))
				(wait-for-breve)
				(dotimes (var num-breve-synchs)
				  (when (= var 50) (execute-breve-command "move-odour-patch-right") (wait-for-breve))
				  (when (= var 125) (execute-breve-command "move-odour-patch-left") (wait-for-breve))
				  (when (= var 290) (execute-breve-command "move-odour-patch-right") (wait-for-breve))
				  (when (= var 300) (execute-breve-command "move-odour-patch-left") (wait-for-breve))
				  (dotimes (var 5)
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-locomotion-controller cns)))
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-lobe cns))))
				  (synchronize-with-breve (central-nervous-system-breve-object cns))
				  ;; Get position, and push it onto the list
				  (push 
				   (execute-breve-command-with-result 
				    "get-location")
				   *position-logs*)
				  ;; Inspect the lobe to get the mem. pot. of
				  ;; right and left bifur neuron
				  (push
				   (ctrnn::neuron-snapshot-firing-frequency
				    (third
				     (neural-network-neurons
				      (carausius-morosus:central-nervous-system-lobe
				       cns))))
				   *left-potentials*)
				  (push
				   (ctrnn::neuron-snapshot-firing-frequency
				    (fourth
				     (neural-network-neurons
				      (carausius-morosus:central-nervous-system-lobe
				       cns))))
				   *right-potentials*)
				  (breve-interface:wait-for-breve))
				;; Return improvement in position
				(let ((new-distance (get-distance-to-target)))
				  (- start-distance new-distance)))))))
       (format t "~A~%" fitnesses)
       (apply #'+ fitnesses))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
				   (fitness genotype num-breve-synchs))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
    (let* ((carausius-morosus::*maximum-leg-force* 200000)
	   (carausius-morosus::*maximum-leg-down-force* 80000)
	   (population 
	    (sort-population (evaluate (make-random-population 100) 600))))
      (with-open-file (f (format nil "logs/turning-without-cc-log~A.txt" id) :direction :output :if-exists :supersede)
	(do ((generation 1 (+ generation 1)))
	    ((> generation num-generations))
	  (format t "*** Generation ~A~%" generation)
	  ;; Select parents
	  (let ((parents (select-parents population 2)))
	    ;; Recombine
	    (let ((children (recombine parents)))
	      ;; Maybe mutate
	      (when (< (random 100) 5)
		(setf children (loop for c in children collect (mutate c))))
	      ;; Evaluate new individuals
	      (setf children (evaluate children 400))
	      ;; Select individuals for next generation
	      (setf population
		    (sort-population (append
				      children
				      (remove-individuals population 2))))
	      ;; Update display
	      (format f "~A ~A ~A ~A ~A~%"
		      generation
		      (cdr (first population)) ; Fitness of best
		      (/ (loop for ind in population
			    sum (cdr ind))
			 (length population))
		      (cdr (first (last population)))
		      (car (first population))) ;; Genome of best
	      (force-output f)))))
      (first population)))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (dotimes (var 10)
    (when (> var 8)
      (ga 250 var))))

(defun test-run (genome &optional (synchs 400))
  (evaluate (list (cons genome 0)) synchs))

(defun test-runs ()
  "Test the 10 genome evolved after 650 generations."
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (test-run
     (vector 10970375 513922 2466507 479636 11395519 264320 12564730 16027703 8428500 16719251 9432762 1449114 7536307 487815 8387234 5080021 3802769 12299044 6530365 13030482 2303181 13139358 1499202 12820760 1232301 7589091 188885 9305580 6444124 12337638 4441937 13863696 1077218 2090969))
    (test-run
     (vector 10957049 513427 2467153 467725 11378291 252852 12577528 16032291 8434417 16714016 9431997 1451770 7531816 492533 8384216 5075164 3806325 12311761 6530705 13025650 2300918 13146603 1490498 12817958 1239940 7945231 11649564 7928140 3550884 11572768 2898581 12328226 6490184 1609901))
    (test-run
     (vector 10971944 521977 2457366 482522 11388430 254725 12565966 16020378 8418631 16706274 9429669 1443941 7522381 501240 8398402 5077330 3797499 12301242 6525757 13024086 2317410 13145069 1489847 12814818 1233111 9887649 3083939 7701638 1319205 14399656 587349 10801335 6689238 533015))
    (test-run
     (vector 10969761 516662 2455267 477192 11383244 263791 12564885 16028260 8429835 16720631 9433119 1441656 7525307 498937 8382771 5079937 3807284 12301649 6531972 13017886 2307373 13133663 1497250 12827499 1228352 9557107 3878634 4846380 3421772 14627416 9325487 15608441 1077543 1740946))
    (test-run
     (vector 10957742 515760 2467769 477431 11393071 252075 12573695 16024981 8416872 16721635 9446247 1438892 7528054 485061 8384796 5069471 3793195 12305459 6528183 13026956 2307312 13136662 1488733 12814506 1224067 8634000 3124438 4065534 3205438 14940404 5644950 15652016 5963209 2769029))
    (test-run
     (vector 10965389 522278 2464779 480694 11378858 257966 12564480 16030900 8435953 16722290 9428684 1451301 7519532 497542 8389406 5081639 3798649 12300061 6538481 13025090 2316738 13143515 1493024 12829175 1240022 8403071 13680348 5152862 120592 14545535 4448716 12994062 8037610 3882983))
    (test-run 
     (vector 10955627 510612 2464954 478581 11377715 265186 12564300 16023917 8426261 16708532 9446987 1453275 7519117 498693 8393546 5083543 3806864 12311746 6522055 13026946 2308718 13122725 1501895 12830156 1224839 8179223 9738398 6305019 3218300 13687115 6927717 11350485 6127392 824351))
    (test-run 
     (vector 10969949 512514 2469081 468275 11385372 255475 12570179 16024822 8420887 16714470 9443172 1454914 7535652 501938 8396633 5085667 3803158 12307250 6535647 13019581 2312251 13128067 1503575 12823374 1242693 8188970 13941506 6961914 5172791 15875871 4627110 14345351 1559489 2703711))
    (test-run 
     (vector 10956562 525138 2450416 478991 11376185 256952 12580733 16016129 8424097 16716414 9446915 1444964 7520039 497047 8380584 5084032 3808183 12305422 6527555 13033917 2307040 13134237 1498087 12813371 1241497 4051738 629603 8004440 527942 15058521 5940884 12919205 404157 124490))
    (test-run
     (vector 10969702 516995 2450403 472760 11375932 253553 12567111 16022128 8425584 16709706 9443753 1448571 7529830 484454 8384214 5071535 3800975 12301345 6520494 13016502 2311673 13146220 1502657 12815948 1237958 8141134 1483562 9300836 7185172 7675544 5579342 11213663 7261428 1146762))))

(defparameter *position-logs* nil) ; holds insect positions
(defparameter *left-potentials* nil) ; hold potentials of left bif. neuron
(defparameter *right-potentials* nil) ; hold potentials of right bif. neuron

(defun test-path ()
  (let* ((carausius-morosus::*maximum-leg-force* 200000)
	 (carausius-morosus::*maximum-leg-down-force* 80000))
    (setf *position-logs* nil
	  *left-potentials* nil
	  *right-potentials* nil)
    (test-run
     (vector 10971944 521977 2457366 482522 11388430 254725 12565966 16020378 8418631 16706274 9429669 1443941 7522381 501240 8398402 5077330 3797499 12301242 6525757 13024086 2317410 13145069 1489847 12814818 1233111 9887649 3083939 7701638 1319205 14399656 587349 10801335 6689238 533015) 1400)
; Better performance, but not resilient     (vector 6707912 519146 5847706 475293 11021734 4034557 13209047 5113294 8388608 1413607 10518733 1691558 7664068 1255015 8388608 2409152 5188667 12542669 6564925 12766363 2653418 13374513 2587161 12544360 1287105 10276338 13811141 2285084 538246 14533227 10895421 11085297 1142809 4815469) 400)
    (with-open-file (log-file "logs/turning-positions.txt" 
			      :if-exists :supersede
			      :direction :output)
      (loop 
	 for position in (reverse *position-logs*)
	 and left in (reverse *left-potentials*)
	 and right in (reverse *right-potentials*)
	 and time from 0
	 do (format log-file "~A ~,2F ~,2F ~A ~A ~A~%" 
		    (* time 50)
		    left 
		    right
		    70
		    (if (or (and (> time 50) (< time 200)) 
			    (and (> time 350) (< time 370)))
			150
			-150)
		    (string-trim '(#\( #\)) position))))))



      

;; Best
;6707912 519146 5847706 475293 11021734 4034557 13209047 5113294 8388608 1413607 10518733 1691558 7664068 1255015 8388608 2409152 5188667 12542669 6564925 12766363 2653418 13374513 2587161 12544360 1287105 10276338 13811141 2285084 538246 14533227 10895421 11085297 1142809 4815469