;;;; What to evolve:
;;;;
;;;; 25 26 27 28:
;;;; - Bias and t.const for right and left sensor neurons
;;;; - Bias and t.const for right and left output neurons
;;;;
;;;; 29 30:
;;;; - Weights from each sensor to outputs
;;;;
;;;; 31:
;;;; - Self weight for output neurons (> 4)
;;;;
;;;; 32:
;;;; - Weights between output neurons
;;;;
;;;; 33:
;;;; - Weight from outputs to leg controller stance neurons (< 0) 
;;;;
;;;; Creature 7

(in-package :navigation)

(setf carausius-morosus::*AEP-L1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R1* (+ (* 1.2 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R2* (+ (* 1.6 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-L3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*AEP-R3* (+ (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R1* (- (* 2.0 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R2* (- (* 2.4 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-L3* (- (* 2.8 (/ pi 4)) 0.01))
(setf carausius-morosus::*PEP-R3* (- (* 2.8 (/ pi 4)) 0.01))

(defun make-random-population (size)
  (loop for n from 1 to size
     collect (let ((individual (loop for gene from 0 to 33
				  collect (cond 
					    ((= gene 31) ; Self weights > 4
					     (make-random-number-between (weight->weight-gene 4) (expt 2 24)))
					    ((= gene 33) ; Negative connection to stance
					     (random (/ (expt 2 24) 2)))
					    (t
					     (random (expt 2 24)))))))
	       ;; Inject earlier evolved genome into the population
	       (setf (subseq individual 0 25)
		     (list
		      10963896 516852 2460398 475135 11385571 260912 12572387 16024549 
		      8426675 16714521 9437184 1445525 7527271 493102 8388608 5077806
		      3802889 12303830 6529969 13024977 2308303 13138285 1497664 12821677
		      1233568))
	       ;; Randomize earlier evolved genome to "neighbourhood" by mutation
	       (loop for n from 0 to 24
		     do (incf (nth n individual) (- (random 20000) 10000)))
	       ;; Ensure individual represent a center crossing net by adjusting bias of outputs to (- (/ sum-of-weights 2))
;	       (setf (nth 27 individual)
;		     (weight->weight-gene
;		      (- (/ (+ (weight-gene->weight (nth 29 individual)) ; from sensor 1
;			       (weight-gene->weight (nth 30 individual)) ; from sensor 2
;			       (weight-gene->weight (nth 31 individual)) ; self weight
;			       (weight-gene->weight (nth 32 individual))) ; from other output
;			    2))))
	       (cons (coerce individual 'vector) 0)))) ; Neither fit nor unfit 

(defun genotype->phenotype (genome)
  (let ((cns (leg-controller::genotype->phenotype 
	      ;; Realize locomotion controller from earlier evolved genome
	      (cons (subseq genome 0 25) 0)))
	;; Add a lobe
	(lobe (make-instance 'neural-network)))
    ;; Add lobe to neural system
    (setf (carausius-morosus:central-nervous-system-lobe cns) lobe)
    ;; Add neurons to the lobe
    (setf (neural-network-neurons lobe)
	  (list (make-instance 'sensor-neuron)
		(make-instance 'sensor-neuron)
		(make-instance 'neuron)
		(make-instance 'neuron)))
    ;; Adjust properties of the neurons
    (setf 
     ;; Gene 25
     (neuron-bias (first (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     (neuron-bias (second (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 25))
     ;; Gene 26
     (neuron-time-constant (first (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     (neuron-time-constant (second (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 26))
     ;; Gene 27
     (neuron-bias (third (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     (neuron-bias (fourth (neural-network-neurons lobe)))
     (weight-gene->weight (svref genome 27))
     ;; Gene 28
     (neuron-time-constant (third (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28))
     (neuron-time-constant (fourth (neural-network-neurons lobe)))
     (time-constant-gene->time-constant (svref genome 28)))
    ;; Add synapses to the output neurons (gene 29 30 31 32)
    (setf (neuron-dendrites (third (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    (setf (neuron-dendrites (fourth (neural-network-neurons lobe)))
	  (list (make-instance 'synapse
			       :from-neuron (second (neural-network-neurons lobe)) ; NOTE: _second_
			       :strength (weight-gene->weight (svref genome 29)))
		(make-instance 'synapse
			       :from-neuron (first (neural-network-neurons lobe)) ; NOTE: then _first_
			       :strength (weight-gene->weight (svref genome 30)))
		(make-instance 'synapse
			       :from-neuron (fourth (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 31)))
		(make-instance 'synapse
			       :from-neuron (third (neural-network-neurons lobe))
			       :strength (weight-gene->weight (svref genome 32)))))
    ;; Add synapses to the middle leg stance neurons (gene 33)
    (push (make-instance 'synapse
			 :from-neuron (third (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites 
	   (leg-controller-stance-neuron 
	    (locomotion-controller-l2
	     (central-nervous-system-locomotion-controller cns)))))
    (push (make-instance 'synapse
			 :from-neuron (fourth (neural-network-neurons lobe))
			 :strength (weight-gene->weight (svref genome 33)))
	  (neuron-dendrites
	   (leg-controller-stance-neuron
	    (locomotion-controller-r2
	     (central-nervous-system-locomotion-controller cns)))))
    ;; Add sensor function to sensor neurons
    (setf (sensor-neuron-sensor-function (first (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (first (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-fol-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (first (neural-network-neurons lobe))))))
	  (sensor-neuron-sensor-function (second (neural-network-neurons lobe)))
	  #'(lambda ()
	      (setf
	       (neuron-external-current (second (neural-network-neurons lobe)))
	       (* (carausius-morosus::carausius-morosus-for-distance (central-nervous-system-breve-object cns))
		  (neuron-external-current-magnitude (second (neural-network-neurons lobe)))))))
    cns))

(defun get-distance-to-target ()
  (read-from-string
   (string-trim
    '(\")
    (execute-breve-command-with-result "get-distance-to-target"))))

(defun get-time ()
  (read-from-string
   (string-trim 
    '(\") (breve-interface:execute-breve-command-with-result "get-time"))))

(defparameter *fitnesses* (make-hash-table :test #'equalp))

(defun fitness (genotype &optional (num-breve-synchs 200))
  (unless (gethash (car genotype) *fitnesses*)
    (setf
     (gethash (car genotype) *fitnesses*)
     ;; Go left first, then right
     (let ((fitnesses
	    (loop for var from 0 to 1
		  collect (progn
			    (breve-interface:execute-breve-command "set-output-filter" -1)
			      (execute-breve-command "reset")
			      (if (= var 0)
				  (execute-breve-command "move-odour-patch-left")
				  (execute-breve-command "move-odour-patch-right"))
			      (wait-for-breve)
			      ;; Start evaluation
			      (let ((energy 0)
				    (cns (genotype->phenotype (car genotype)))
				    (start-distance (get-distance-to-target)))
				(dotimes (var num-breve-synchs)
				  (dotimes (var 5)
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-locomotion-controller cns)))
				    (incf energy
					  (synchronously-update-membrane-potentials!
					   (central-nervous-system-lobe cns))))
				  (synchronize-with-breve (central-nervous-system-breve-object cns))
				  (breve-interface:wait-for-breve))
				;; Return improvement in position
				(let ((new-distance (get-distance-to-target)))
				  (- start-distance new-distance)))))))
       (format t "~A~%" fitnesses)
       (apply #'+ fitnesses))))
  (gethash (car genotype) *fitnesses*))

(defun evaluate (individuals &optional (num-breve-synchs 200))
  "Evaluate genotypes and return the list of evaluated genotypes."
  (loop for genotype in individuals
        and num from 1
        collect (progn
                  (format t "~&evaluating ~A (~A of ~A)...~%"
                          (car genotype)
                          num
                          (length individuals))
                  (let ((res (cons (car genotype)
				   (fitness genotype num-breve-synchs))))
                    (format t "~A~%" (cdr res))
                    res))))

(defun sort-population (population)
  "Sort population from best to worst fitness."
  (sort (copy-list population)
        #'(lambda (one two)
            (> (cdr one) (cdr two)))))

(defun remove-individuals (population &optional (num-to-slay 2))
  "Select two individuals to slay, from worst quartile of population."
  (let ((population (copy-list population)))
    (butlast population num-to-slay)))

(defun select-parents (population &optional (num-parents 2))
  "Pick parents from best quartile of population."
  (let ((population (copy-list population)))
    (loop for parent-num from 1 to num-parents
          collect (let ((parent
                         (nth (make-random-number-between
                               0
                               (- (round (/ (length population) 4)) parent-num))
                              population)))
                    (setf population (remove parent population))
                    parent))))

(defun recombine (parents)
  "Recombine with two-point crossover the two first individuals in 'parents'."
  (let* ((p1 (first parents))
         (p2 (second parents))
         (c1 (copy-seq (car p1)))
         (c2 (copy-seq (car p2)))
         (y (make-random-number-between 0 (- (length c1) 2)))
         (x (make-random-number-between y (- (length c1) 1))))
    (psetf (subseq c1 y x) (subseq (car p2) y x)
           (subseq c2 y x) (subseq (car p1) y x))
    (list (cons c1 0) (cons c2 0))))

(defun mutate (genotype)
  "Return a mutation of the given genotype."
  (let ((x (make-random-number-between 0 (- (length (car genotype)) 1)))
        (new-genome (copy-seq (car genotype)))
        (rand (- (random 200000) 100000)))
    ;; If current value is within the range of the number to add/subtract
    (when (and (< (aref new-genome x) (- (expt 2 24) rand))
               (> (aref new-genome x) rand))
      (incf (aref new-genome x) rand))
    (cons new-genome (cdr genotype))))

(defun ga (num-generations &optional (id 1))
    (let* ((carausius-morosus::*maximum-leg-force* 200000)
	   (carausius-morosus::*maximum-leg-down-force* 80000)
	   (population 
	    (sort-population (evaluate (make-random-population 100) 400))))
      (with-open-file (f (format nil "logs/turning-without-cc-log~A.txt" id) :direction :output :if-exists :supersede)
	(do ((generation 1 (+ generation 1)))
	    ((> generation num-generations))
	  (format t "*** Generation ~A~%" generation)
	  ;; Select parents
	  (let ((parents (select-parents population 2)))
	    ;; Recombine
	    (let ((children (recombine parents)))
	      ;; Maybe mutate
	      (when (< (random 100) 5)
		(setf children (loop for c in children collect (mutate c))))
	      ;; Evaluate new individuals
	      (setf children (evaluate children 400))
	      ;; Select individuals for next generation
	      (setf population
		    (sort-population (append
				      children
				      (remove-individuals population 2))))
	      ;; Update display
	      (format f "~A ~A ~A ~A ~A~%"
		      generation
		      (cdr (first population)) ; Fitness of best
		      (/ (loop for ind in population
			    sum (cdr ind))
			 (length population))
		      (cdr (first (last population)))
		      (car (first population))) ;; Genome of best
	      (force-output f)))))
      (first population)))

(defun 10-runs ()
  (setf *fitnesses* (make-hash-table :test #'equalp))
  (dotimes (var 10)
    (when (> var 8)
      (ga 250 var))))

(defun test-run-good ()
  (let ((carausius-morosus::*maximum-leg-force* 200000)
	(carausius-morosus::*maximum-leg-down-force* 80000))
    (fitness
     (cons
      ;; (vector 10971944 521977 2457366 482522 11388430 254725 12565966 16020378 8418631 16706274 9429669 1443941 7522381 501240 8398402 5077330 3797499 12301242 6525757 13024086 2317410 13145069 1489847 12814818 1233111 9887649 3083939 7701638 1319205 14399656 587349 10801335 6689238 533015) ; best incremental
     (vector
      6707912 519146 5847706 475293 11021734 4034557 13209047 5113294 8388608 1413607 10518733 1691558 7664068 1255015 8388608 2409152 5188667 12542669 6471289 12766363 2653418 13374513 2587161 12544360 1287105 10276338 13811141 2285084 538246 14533227 10895421 11085297 1142809 4815469)
     0)
    400)))
    





      

