;; Utilites common to all of the evolutionary algorithms code

(defpackage #:utilities
  (:use "COMMON-LISP" "CTRNN")
  (:export
   #:get-location
   #:make-random-number-between
   #:make-random-number-not-in
   #:weight-gene->weight
   #:time-constant-gene->time-constant
   #:weight->weight-gene))

(in-package :utilities)

(defun weight-gene->weight (gene)
  "Convert 24-bit integer into a real weight value."
  (+ -20 (/ gene (/ (expt 2 24) 40))))

(defun time-constant-gene->time-constant (gene)
  (+ (/ gene (/ (expt 2 24) 16)) ctrnn::*timestep*))

(defun weight->weight-gene (weight)
  "Convert real weight value into 24-bit integer."
  (round (* (+ weight 20) (/ (expt 2 24) 40))))

(defun make-random-number-between (min max &optional (decimals 0))
  "Create a random number between min and max (inclusive)."
  (if (or (> min max))
    0
    (let ((max (* (expt 10 decimals) max))
          (min (* (expt 10 decimals) min)))
      (/ (+ min (random (- max min -1))) (expt 10 decimals)))))

(defun make-random-number-not-in (max &optional list-of-numbers)
  "Make a random number that is not in list-of-numbers."
  (loop for a = (random max)
        when (not (member a list-of-numbers))
        return a))

(defun get-location ()
  "Get the location of the breve creature."
  (car
   (read-from-string
    (coerce
     (loop for char across
           (breve-interface:execute-breve-command-with-result "get-location")
           unless (char-equal #\, char)
           collect char) 'string))))
