(in-package :ctrnn)

(defun test-neuron ()
  (let ((neuron (make-instance 'neuron :membrane-potential 0 :time-constant 1.0))
        (plot-pane (make-instance 'capi-extensions::plot-pane
                                  :x-min 0 :x-max 10 :x-axis 1
                                  :y-min -10 :y-max 10 :y-axis 0.1))
        (plot-pane-graph (make-instance 'capi-extensions::plot-pane-graph :colour :blue)))
    (push plot-pane-graph (capi-extensions::plot-pane-graphs plot-pane))
    ;; Integrate neurons for five seconds
    (dotimes (var 100)
      (let ((*timestep* 0.1)
            (ext (cond ((< var 20)
                        5)
                       (t 1))))
        (setf (neuron-external-current neuron) ext)
        (push (cons (/ var 10) (neuron-membrane-potential neuron))
              (capi-extensions::plot-pane-graph-points plot-pane-graph))
        (update-membrane-potential! neuron)))
    (capi:contain plot-pane)))

;(test-neuron) 

(defun test-net ()
  (let ((ctrnn (make-instance 'neural-network
                              :neurons (list (make-instance 'neuron :membrane-potential 1)
                                             (make-instance 'neuron :membrane-potential 1)
                                             (make-instance 'neuron :membrane-potential 1)))))
    (add-synapse! (nth 0 (neural-network-neurons ctrnn))
                  (make-instance 'synapse
                                 :from-neuron (nth 0 (neural-network-neurons ctrnn))
                                 :strength 1.0))
    (dolist (neuron (neural-network-neurons ctrnn))
      (format t "neuron pot: ~A~%" (neuron-membrane-potential neuron)))
    (dotimes (var 20)
      (synchronously-update-membrane-potentials! ctrnn))
    (dolist (neuron (neural-network-neurons ctrnn))
      (format t "neuron pot: ~A~%" (neuron-membrane-potential neuron)))))

(defun test-pulse-net ()
  (let ((ctrnn (make-instance 'neural-network
                              :neurons (list (make-instance 'neuron
                                                            :membrane-potential 0
                                                            :time-constant 0.3
                                                            :bias -2)
                                             (make-instance 'neuron
                                                            :membrane-potential 0
                                                            :time-constant 0.1
                                                            :bias 0))))
        (plot-pane (make-instance 'capi-extensions::plot-pane
                                  :x-min -100 :x-max 5000 :x-axis 1000
                                  :y-min -10 :y-max 10 :y-axis 1))
        (n-1-graph (make-instance 'capi-extensions::plot-pane-graph :colour :red))
        (n-2-graph (make-instance 'capi-extensions::plot-pane-graph :colour :blue)))
    (push n-1-graph (capi-extensions::plot-pane-graphs plot-pane))
    (push n-2-graph (capi-extensions::plot-pane-graphs plot-pane))
    (capi:contain plot-pane)
    (add-synapse! (nth 0 (neural-network-neurons ctrnn))
                  (make-instance 'synapse
                                 :from-neuron (nth 1 (neural-network-neurons ctrnn))
                                 :strength 1.0))
    (add-synapse! (nth 1 (neural-network-neurons ctrnn))
                  (make-instance 'synapse
                                 :from-neuron (nth 0 (neural-network-neurons ctrnn))
                                 :strength 1.0))
    (add-synapse! (nth 0 (neural-network-neurons ctrnn))
                  (make-instance 'synapse
                                 :from-neuron (nth 0 (neural-network-neurons ctrnn))
                                 :strength -30.0 ))
    (add-synapse! (nth 1 (neural-network-neurons ctrnn))
                  (make-instance 'synapse
                                 :from-neuron (nth 1 (neural-network-neurons ctrnn))
                                 :strength 4.0))
    (do ((x 0 (+ x 10)))
        ((> x 5000))
      (let ((ctrnn::*timestep* 0.01))
        (synchronously-update-membrane-potentials! ctrnn))
      (push (cons x (neuron-membrane-potential (nth 0 (neural-network-neurons ctrnn))))
            (capi-extensions::plot-pane-graph-points n-1-graph))
      (push (cons x (neuron-membrane-potential (nth 1 (neural-network-neurons ctrnn))))
            (capi-extensions::plot-pane-graph-points n-2-graph))
      (gp:invalidate-rectangle plot-pane))))
