(in-package :ctrnn)

(defparameter *timestep* 0.01 "Integration timestep in seconds.")

(defclass neural-network ()
  ((neurons
    :accessor neural-network-neurons
    :initarg :neurons
    :initform nil))
  (:documentation "A network containing a bunch of neurons."))

(defmethod synchronously-update-membrane-potentials! ((neural-network neural-network))
  "(Destructively) update membrane potentials in the network. Return consumed energy."
  ;; First, calculate all firing frequency at this point in time, and store them
  (loop for neuron in (neural-network-neurons neural-network)
        do (setf (neuron-snapshot-firing-frequency neuron)
                 (firing-frequency neuron)))
  ;; Now, using the snapshot frequency, update all membrane potentials
  (loop for neuron in (neural-network-neurons neural-network)
        sum (let ((energy (update-membrane-potential! neuron)))
              (when (< energy 0)
                (format t "BUG: Negative energy consumption! Neurons _producing_ energy?"))
              energy)))
